@@ -16,6 +16,7 @@
 
 #include <errno.h>
 #include <sched.h>
+#include <stdlib.h>
 #include <unistd.h>
 #include "pthread.h"
 #include "internals.h"
@@ -37,8 +38,6 @@ void pthread_exit(void * retval)
   /* Store return value */
   __pthread_lock(THREAD_GETMEM(self, p_lock), self);
   THREAD_SETMEM(self, p_retval, retval);
-  /* Say that we've terminated */
-  THREAD_SETMEM(self, p_terminated, 1);
   /* See whether we have to signal the death.  */
   if (THREAD_GETMEM(self, p_report_events))
     {
@@ -47,8 +46,8 @@ void pthread_exit(void * retval)
       uint32_t mask = __td_eventmask (TD_DEATH);
 
       if ((mask & (__pthread_threads_events.event_bits[idx]
-		   | THREAD_GETMEM(self,
-				   p_eventbuf.eventmask).event_bits[idx]))
+		   | THREAD_GETMEM_NC(self,
+				      p_eventbuf.eventmask.event_bits[idx])))
 	  != 0)
 	{
 	  /* Yep, we have to signal the death.  */
@@ -60,6 +59,8 @@ void pthread_exit(void * retval)
 	  __linuxthreads_death_event();
 	}
     }
+  /* Say that we've terminated */
+  THREAD_SETMEM(self, p_terminated, 1);
   /* See if someone is joining on us */
   joining = THREAD_GETMEM(self, p_joining);
   __pthread_unlock(THREAD_GETMEM(self, p_lock));
@@ -73,9 +74,13 @@ void pthread_exit(void * retval)
     request.req_kind = REQ_MAIN_THREAD_EXIT;
     __libc_write(__pthread_manager_request, (char *)&request, sizeof(request));
     suspend(self);
+    /* Main thread flushes stdio streams and runs atexit functions.
+       It also calls a handler within LinuxThreads which sends a process exit
+       request to the thread manager. */
+    exit(0);
   }
-  /* Exit the process (but don't flush stdio streams, and don't run
-     atexit functions). */
+  /* Threads other than the main one  terminate without flushing stdio streams
+     or running atexit functions. */
   _exit(0);
 }
 
@@ -112,7 +117,7 @@ int pthread_join(pthread_t thread_id, void ** thread_return)
   extr.pu_extricate_func = join_extricate_func;
 
   __pthread_lock(&handle->h_lock, self);
-  if (invalid_handle(handle, thread_id)) {
+  if (nonexisting_handle(handle, thread_id)) {
     __pthread_unlock(&handle->h_lock);
     return ESRCH;
   }
@@ -129,7 +134,7 @@ int pthread_join(pthread_t thread_id, void ** thread_return)
   /* If not terminated yet, suspend ourselves. */
   if (! th->p_terminated) {
     /* Register extrication interface */
-    __pthread_set_own_extricate_if(self, &extr); 
+    __pthread_set_own_extricate_if(self, &extr);
     if (!(THREAD_GETMEM(self, p_canceled)
 	&& THREAD_GETMEM(self, p_cancelstate) == PTHREAD_CANCEL_ENABLE))
       th->p_joining = self;
@@ -138,13 +143,13 @@ int pthread_join(pthread_t thread_id, void ** thread_return)
     __pthread_unlock(&handle->h_lock);
 
     if (already_canceled) {
-      __pthread_set_own_extricate_if(self, 0); 
+      __pthread_set_own_extricate_if(self, 0);
       pthread_exit(PTHREAD_CANCELED);
     }
 
     suspend(self);
     /* Deregister extrication interface */
-    __pthread_set_own_extricate_if(self, 0); 
+    __pthread_set_own_extricate_if(self, 0);
 
     /* This is a cancellation point */
     if (THREAD_GETMEM(self, p_woken_by_cancel)
